import cv2
import numpy as np


def roundness(contour: np.ndarray) -> float:
    """
    Calculate the roundness of a contour.

    The roundness is determined by the ratio of the area of the contour to the area of
    the minimum enclosing circle.

    Args:
        contour (np.ndarray): An OpenCV contour of a single object.

    Returns:
        float: The roundness value, ranging from 0 to 1, where 1 indicates a perfect circle.

    Examples:
        >>> contour = np.array([[[1, 2]], [[2, 3]], [[3, 2]], [[2, 1]]])
        >>> roundness(contour)
        0.8367346938775511
    """

    area = cv2.contourArea(contour)
    _, radius = cv2.minEnclosingCircle(contour)
    enclosing_circle_area = radius**2 * np.pi
    roundness = area / enclosing_circle_area
    return roundness



from unittest import TestCase
from .main import overlap_object


class Test_overlap_object(TestCase):
    def test_ellipse(self):
        object1 = {
            "type": "ellipse",
            "center": {"x": 4, "y": 4},
            "radiusX": 3,
            "radiusY": 3,
            "angleOfRotation": 0.0,
        }
        object2 = {
            "type": "ellipse",
            "center": {"x": 6, "y": 4},
            "radiusX": 7,
            "radiusY": 3,
            "angleOfRotation": 50,
        }

        overlap = overlap_object(object1, object2)
        expOverlap = 0.772
        self.assertAlmostEqual(overlap, expOverlap, places=1)

    def test_completeOverlap(self):
        object1 = {
            "type": "ellipse",
            "center": {"x": 30000, "y": 5000},
            "radiusX": 7,
            "radiusY": 3,
            "angleOfRotation": 0.0,
        }
        object2 = {
            "type": "ellipse",
            "center": {"x": 30000, "y": 5000},
            "radiusX": 3000,
            "radiusY": 2000,
            "angleOfRotation": 50,
        }

        overlap = overlap_object(object1, object2)
        expOverlap = 1
        self.assertAlmostEqual(overlap, expOverlap, places=1)

    def test_noOverlap(self):
        object1 = {
            "type": "ellipse",
            "center": {"x": 30000, "y": 5000},
            "radiusX": 7,
            "radiusY": 3,
            "angleOfRotation": 0.0,
        }
        object2 = {
            "type": "ellipse",
            "center": {"x": 3, "y": 5},
            "radiusX": 3,
            "radiusY": 20,
            "angleOfRotation": 50,
        }

        overlap = overlap_object(
            object1,
            object2,
        )
        expOverlap = 0
        self.assertAlmostEqual(overlap, expOverlap, places=1)

    def test_noArea(self):
        object1 = {
            "type": "ellipse",
            "center": {"x": 30000, "y": 5000},
            "radiusX": 0,
            "radiusY": 0,
            "angleOfRotation": 0.0,
        }
        object2 = {
            "type": "ellipse",
            "center": {"x": 3000, "y": 5000},
            "radiusX": 0,
            "radiusY": 0,
            "angleOfRotation": 0,
        }

        overlap = overlap_object(object1, object2)
        expOverlap = 0
        self.assertAlmostEqual(overlap, expOverlap, places=1)

    def test_polypon(self):
        object1 = {
            "type": "polygon",
            "points": [
                {"x": 2, "y": 3},
                {"x": 5, "y": 6},
                {"x": 2, "y": 5},
                {"x": 7, "y": 9},
                {"x": 4, "y": 2},
            ],
        }
        object2 = {
            "type": "polygon",
            "points": [
                {"x": 5, "y": 3},
                {"x": 10, "y": 10},
                {"x": 4, "y": 5},
                {"x": 4, "y": 2},
            ],
        }

        overlap = overlap_object(object1, object2)
        expOverlap = 0.286
        self.assertAlmostEqual(overlap, expOverlap, places=1)

    def test_polygon_ellipse(self):
        object1 = {
            "type": "polygon",
            "points": [
                {"x": 2, "y": 3},
                {"x": 5, "y": 6},
                {"x": 2, "y": 5},
                {"x": 7, "y": 9},
                {"x": 4, "y": 2},
            ],
        }
        object2 = {
            "type": "ellipse",
            "center": {"x": 6, "y": 4},
            "radiusX": 7,
            "radiusY": 3,
            "angleOfRotation": 50,
        }

        overlap = overlap_object(object1, object2)
        expOverlap = 0.899
        self.assertAlmostEqual(overlap, expOverlap, places=1)

    def test_ellipse_polygon(self):
        object1 = {
            "type": "ellipse",
            "center": {"x": 4, "y": 4},
            "radiusX": 3,
            "radiusY": 3,
            "angleOfRotation": 0.0,
        }

        object2 = {
            "type": "polygon",
            "points": [
                {"x": 5, "y": 3},
                {"x": 10, "y": 10},
                {"x": 4, "y": 5},
                {"x": 4, "y": 2},
            ],
        }

        overlap = overlap_object(object1, object2)
        expOverlap = 0.215
        self.assertAlmostEqual(overlap, expOverlap, places=1)

    def test_unknown_type(self):
        object1 = {"type": "square", "x1": 4, "y1": 4, "x2": 3, "y2": 3}
        object2 = {
            "type": "ellipse",
            "center": {"x": 6, "y": 4},
            "radiusX": 7,
            "radiusY": 3,
            "angleOfRotation": 50,
        }
        self.assertRaises(ValueError, overlap_object, object1, object2)

    def test_negative(self):
        object1 = {
            "type": "ellipse",
            "center": {"x": 4, "y": 4},
            "radiusX": -1,
            "radiusY": 3,
            "angleOfRotation": 0.0,
        }
        object2 = {
            "type": "ellipse",
            "center": {"x": 6, "y": 4},
            "radiusX": 7,
            "radiusY": 3,
            "angleOfRotation": 50,
        }

        self.assertRaises(ValueError, overlap_object, object1, object2)

    def test_bug_to_small_object(self):
        object1 = {
            "type": "polygon",
            "points": [
                {"x": 618, "y": 54},
                {"x": 625, "y": 61},
            ],
            "id": "56acfd7c-de50-4c45-906f-d97b4ea6467d",
            "parents": [],
            "children": [],
        }
        object2 = {
            "type": "polygon",
            "points": [
                {"x": 618, "y": 54},
                {"x": 618, "y": 61},
                {"x": 625, "y": 61},
                {"x": 625, "y": 54},
            ],
            "id": "56acfd7c-de50-4c45-906f-d97b4ea6467d",
            "parents": [],
            "children": [],
        }

        overlap = overlap_object(object1, object2)
        expOverlap = 0
        self.assertAlmostEqual(overlap, expOverlap, places=1)



import json
import cv2
from tkinter import filedialog

from tomni.transformers import mask2json

file = filedialog.askopenfilename()
mask = cv2.imread(file, 0)
json_objects = mask2json(mask)

json_objects = [
    x for i, x in enumerate(json_objects) if len(json_objects[i]["points"]) >= 3
]

with open(file.replace(".png", ".json"), "w") as f:
    json.dump(json_objects, f)



from typing import List, Union
import copy
from ...transformers import json2contours
from ...contour_operations import get_center
from ..translation import translation_json


def json_object_to_keep(
    json_object: dict, new_x: tuple, new_y: tuple, crop_mode: str = "remove_objects"
) -> bool:
    """
    Check if a JSON object should be kept or removed based on its position relative to the cropped area.

    Args:
        json_object (dict): JSON object following the standard AxionBio format.
        new_x (tuple): Tuple (xmin_crop, xmax_crop) representing the x dimensions of the cropped image.
        new_y (tuple): Tuple (ymin_crop, ymax_crop) representing the y dimensions of the cropped image.
        crop_mode (str): Crop mode, either "remove_objects" (default) or "keep_objects".

    Returns:
        bool: True if the JSON object should be kept, False otherwise.

    Note:
        - In "remove_objects" mode, the JSON object is removed if its center is outside the crop area.
        - In "keep_objects" mode, JSON objects with centers outside the crop but with parts inside the
          crop are kept, and new JSON objects are created with adjusted centers.

    Raises:
        ValueError: If the type of the annotation in the JSON object is not "polygon" or "ellipse".
    """
    if crop_mode == "remove_objects":
        x, y = 0, 0
        if json_object["type"] == "polygon":
            contour = json2contours(json_object)
            x, y = get_center(contour)
        elif json_object["type"] == "ellipse":
            x, y = json_object["center"]["x"], json_object["center"]["y"]
        else:
            raise ValueError(f"No json type found of {json_object['type']}")

        within_crop = True
        if x < new_x[0] or x > new_x[1] or y < new_y[0] or y > new_y[1]:
            within_crop = False
    else:
        within_crop = True

    return within_crop


def crop_json(
    json_list: List[dict],
    x_translation: int,
    y_translation: int,
    crop_dim: Union[list, tuple],
    crop_mode: str = "remove_objects",
) -> List[dict]:
    """
    Crop a list of JSON objects based on translation and crop dimensions.

    Args:
        json_list (List[dict]): A list of JSON objects in standard AxionBio format.
        x_translation (int): The translation in the x-direction of the cropped JSON relative to the original JSON.
        y_translation (int): The translation in the y-direction of the cropped JSON relative to the original JSON.
        crop_dim (Union[list, tuple]): Crop dimensions (x, y) of the cropped JSON.
        crop_mode (str): Crop mode, either "remove_objects" (default) or "keep_objects".

    Returns:
        List[dict]: A list of JSON objects after cropping and translation.

    Note:
        - In "remove_objects" mode, JSON objects with centers outside the crop area are removed.
        - In "keep_objects" mode, JSON objects with centers outside the crop area but with parts inside the
          crop area are kept, and new JSON objects are created with adjusted centers.
    """

    new_json_list = []

    json_list_copy = copy.deepcopy(json_list)

    for c, json_object in enumerate(json_list):
        json_copy = copy.deepcopy(json_object)
        to_keep = json_object_to_keep(
            json_copy,
            (x_translation, x_translation + crop_dim[0]),
            (y_translation, y_translation + crop_dim[1]),
            crop_mode,
        )
        if to_keep:
            json_list_copy[c] = translation_json(
                json_copy, -x_translation, -y_translation
            )

        else:
            continue

        new_json_list.append(json_list_copy[c])

    return new_json_list



from typing import Union
import numpy as np
from ..contours2json import contours2json


def list_of_points2json(list_of_points: Union[list, np.array]) -> dict:
    """
    Convert a list of points into a JSON object representing a polygon.

    Args:
        list_of_points (Union[list, np.ndarray]): The list of points describing a polygon in the form of
            a list or NumPy array, where each point is represented as [x, y].

    Returns:
        dict: A JSON object representing a polygon.

    Example::

        # Define a list of points representing a polygon
        polygon_points = [[1, 1], [2, 1], [2, 2], [1, 2]]

        # Convert the list of points to a JSON object
        polygon_json = list_of_points_to_json(polygon_points)

        # Print the resulting JSON object
        polygon_json = {
            'type': 'polygon',
            'points': [{'x': 1, 'y': 1}, {'x': 2, 'y': 1}, {'x': 2, 'y': 2}, {'x': 1, 'y': 2}]
        }


    Note:
        - The input `list_of_points` should be a list or NumPy array where each element is a point represented as [x, y].
        - The resulting JSON object represents a polygon with type 'polygon' and a list of points.

    """
    contours = np.array(list_of_points)
    contours = contours.reshape((1, contours.shape[0], 1, contours.shape[1]))

    json_object = contours2json(contours)[0]
    return json_object



from unittest import TestCase

import numpy as np

from tomni.annotation_manager.annotations.point import Point

from .main import parse_points_to_contour


class TestUtils(TestCase):
    def setUp(self) -> None:
        self.epsilon = 0.9
        self.n_iter = 1
        self.circular_points = [
            Point(1, 2),
            Point(1, 4),
            Point(2, 5),
            Point(4, 5),
            Point(5, 4),
            Point(5, 2),
            Point(4, 1),
            Point(3, 1),
            Point(2, 1),
        ]
        self.star_shaped_points = [
            Point(1, 3),
            Point(2, 3),
            Point(3, 5),
            Point(5, 3),
            Point(3, 1),
            Point(2, 2),
        ]
        self.rectangle_points = [
            Point(1, 5),
            Point(3, 5),
            Point(5, 5),
            Point(5, 1),
            Point(3, 1),
            Point(1, 1),
        ]
        self.triangle_points = [
            Point(1, 5),
            Point(3, 1),
            Point(5, 5),
        ]

    def tearDown(self):
        pass

    def test_parse_points_to_contour_star(self):
        expected = np.array(
            [[[1, 3]], [[2, 3]], [[3, 5]], [[5, 3]], [[3, 1]], [[2, 2]]], dtype=np.int32
        )

        actual = parse_points_to_contour(self.star_shaped_points)

        np.testing.assert_array_equal(expected, actual)
        self.assertEqual(expected.dtype, actual.dtype)



from unittest import TestCase

from .main import iterative_downsampling


class TestIterativeDownsampling(TestCase):
    def setUp(self) -> None:
        self.n_iter = 1
        self.circular_points = [
            {"x": 1, "y": 2},
            {"x": 1, "y": 4},
            {"x": 2, "y": 5},
            {"x": 4, "y": 5},
            {"x": 5, "y": 4},
            {"x": 5, "y": 2},
            {"x": 4, "y": 1},
            {"x": 3, "y": 1},
            {"x": 2, "y": 1},
        ]
        self.star_shaped_points = [
            {"x": 1, "y": 3},
            {"x": 2, "y": 3},
            {"x": 3, "y": 5},
            {"x": 5, "y": 3},
            {"x": 3, "y": 1},
            {"x": 2, "y": 2},
        ]
        self.rectangle_points = [
            {"x": 1, "y": 5},
            {"x": 3, "y": 5},
            {"x": 5, "y": 5},
            {"x": 5, "y": 1},
            {"x": 3, "y": 1},
            {"x": 1, "y": 1},
        ]
        self.triangle_points = [
            {"x": 1, "y": 5},
            {"x": 3, "y": 1},
            {"x": 5, "y": 5},
        ]

    def test_iterative_downsampling_rectangle(self):
        expected = [{"x": 1, "y": 5}, {"x": 5, "y": 5}, {"x": 3, "y": 1}]

        actual = iterative_downsampling(self.rectangle_points, self.n_iter)

        self.assertEqual(expected, actual)

    def test_iterative_downsampling_star(self):
        expected = [{"x": 1, "y": 3}, {"x": 3, "y": 5}, {"x": 3, "y": 1}]

        actual = iterative_downsampling(self.star_shaped_points, self.n_iter)

        self.assertEqual(expected, actual)

    def test_iterative_downsampling_circular(self):
        expected = [
            {"x": 1, "y": 2},
            {"x": 2, "y": 5},
            {"x": 5, "y": 4},
            {"x": 4, "y": 1},
            {"x": 2, "y": 1},
        ]

        actual = iterative_downsampling(self.circular_points, self.n_iter)

        self.assertEqual(expected, actual)

    def test_iterative_downsampling_zero(self):
        actual = iterative_downsampling(self.circular_points, 0)
        self.assertEqual(self.circular_points, actual)

    def test_raises_assertion(self):
        self.assertRaises(AssertionError, iterative_downsampling, [], -1)



import math
from unittest import TestCase

from .main import add_area


class TestAddAreaJson(TestCase):
    def test_happy_flow(self):
        json_object = {
            "type": "polygon",
            "points": [
                {"x": 0, "y": 10},
                {"x": 10, "y": 10},
                {"x": 10, "y": 0},
                {"x": 0, "y": 0},
            ],
            "id": "unicorn",
            "parents": [],
            "children": [],
        }

        expected_result = {
            "type": "polygon",
            "points": [
                {"x": 0, "y": 10},
                {"x": 10, "y": 10},
                {"x": 10, "y": 0},
                {"x": 0, "y": 0},
            ],
            "id": "unicorn",
            "parents": [],
            "children": [],
            "area": 100.0,
        }

        add_area(json_object)

        self.assertDictEqual(json_object, expected_result)

    def test_happy_flow_inner_contours(self):
        json_object = {
            "type": "polygon",
            "points": [
                {"x": 0, "y": 10},
                {"x": 10, "y": 10},
                {"x": 10, "y": 0},
                {"x": 0, "y": 0},
            ],
            "id": "unicorn",
            "parents": [],
            "children": [],
            "inner_points": [
                [
                    {"x": 0, "y": 3},
                    {"x": 3, "y": 3},
                    {"x": 3, "y": 0},
                    {"x": 0, "y": 0},
                ],
                [
                    {"x": 0, "y": 2},
                    {"x": 2, "y": 2},
                    {"x": 2, "y": 0},
                    {"x": 0, "y": 0},
                ],
            ],
        }

        expected_result = {
            "type": "polygon",
            "points": [
                {"x": 0, "y": 10},
                {"x": 10, "y": 10},
                {"x": 10, "y": 0},
                {"x": 0, "y": 0},
            ],
            "id": "unicorn",
            "parents": [],
            "children": [],
            "inner_points": [
                [
                    {"x": 0, "y": 3},
                    {"x": 3, "y": 3},
                    {"x": 3, "y": 0},
                    {"x": 0, "y": 0},
                ],
                [
                    {"x": 0, "y": 2},
                    {"x": 2, "y": 2},
                    {"x": 2, "y": 0},
                    {"x": 0, "y": 0},
                ],
            ],
            "area": 87.0,
        }

        add_area(json_object)

        self.assertDictEqual(json_object, expected_result)

    def test_happy_flow_inner_contours_empty(self):
        json_object = {
            "type": "polygon",
            "points": [
                {"x": 0, "y": 10},
                {"x": 10, "y": 10},
                {"x": 10, "y": 0},
                {"x": 0, "y": 0},
            ],
            "id": "unicorn",
            "parents": [],
            "children": [],
            "inner_points": [],
        }

        expected_result = {
            "type": "polygon",
            "points": [
                {"x": 0, "y": 10},
                {"x": 10, "y": 10},
                {"x": 10, "y": 0},
                {"x": 0, "y": 0},
            ],
            "id": "unicorn",
            "parents": [],
            "children": [],
            "inner_points": [],
            "area": 100.0,
        }

        add_area(json_object)

        self.assertDictEqual(json_object, expected_result)

    def test_circle(self):
        json_object = {
            "type": "circle",
            "center": {"x": 20, "y": 30},
            "radiusX": 10,
            "radiusY": 10,
            "angleOfRotation": 0,
            "id": "unicorn",
        }

        expected_result = {
            "type": "circle",
            "center": {"x": 20, "y": 30},
            "radiusX": 10,
            "radiusY": 10,
            "angleOfRotation": 0,
            "id": "unicorn",
            "area": math.pi * 10 * 10,
        }

        add_area(json_object)

        self.assertDictEqual(json_object, expected_result)

    def test_ellipse(self):
        json_object = {
            "type": "ellipse",
            "center": {"x": 20, "y": 30},
            "radiusX": 10,
            "radiusY": 12,
            "angleOfRotation": 0,
            "id": "unicorn",
        }

        expected_result = {
            "type": "ellipse",
            "center": {"x": 20, "y": 30},
            "radiusX": 10,
            "radiusY": 12,
            "angleOfRotation": 0,
            "id": "unicorn",
            "area": math.pi * 10 * 12,
        }

        add_area(json_object)

        self.assertDictEqual(json_object, expected_result)

    def test_ellipse_overwrite(self):
        json_object = {
            "type": "ellipse",
            "center": {"x": 20, "y": 30},
            "radiusX": 10,
            "radiusY": 12,
            "angleOfRotation": 0,
            "id": "unicorn",
            "area": math.pi * 10 * 12,
        }

        expected_result = {
            "type": "ellipse",
            "center": {"x": 20, "y": 30},
            "radiusX": 10,
            "radiusY": 12,
            "angleOfRotation": 0,
            "id": "unicorn",
            "area": math.pi * 10 * 12,
        }

        add_area(json_object)

        self.assertDictEqual(json_object, expected_result)

    def test_value_error(self):
        json_object = {
            "type": "elipse",
            "center": {"x": 20, "y": 30},
            "radiusX": 10,
            "radiusY": 12,
            "angleOfRotation": 0,
            "id": "unicorn",
        }
        self.assertRaises(ValueError, add_area, json_object)



import numpy as np

from ..labels2listsOfPoints import labels2listsOfPoints
from ..positions2contour import positions2contour


def labels2contours(
    labels: np.ndarray, simplify_error: float = 0, return_inner_contours: bool = False
) -> list:
    """
    Convert a labeled image into a list of OpenCV contours.

    Args:
        labels (np.ndarray): An array where every pixel is labeled to indicate which object it belongs to.
            The labels should be non-negative integers.
        return_inner_contours (bool, optional): If True, return the internal contours (contours around holes within objects).
            If False (default), return only the external contours (outlines of objects).

    Returns:
        List[np.ndarray]: A list of OpenCV-style contours represented as NumPy arrays. Each contour is a 2D array
            with shape (N, 1, 2), where N is the number of points in the contour, and each point has (x, y) coordinates.
            The list contains contours for each labeled object in the input image.

    Warnings:
        - The parameter 'simplify_error' has been deprecated and is no longer used.
          Please update your code to remove references to 'simplify_error'.
    """
    listPoints = labels2listsOfPoints(labels)
    contours = [
        positions2contour(
            p,
            simplify_error=simplify_error,
            return_inner_contours=return_inner_contours,
        )
        for p in listPoints
        if len(p) > 0
    ]
    return contours



def translation_json(json_object: dict, x_translation: int, y_translation: int):
    """
    Translate a JSON object (ellipse or polygon) by specified amounts in the x and y directions.

    Args:
        json_object (dict): A JSON object following the standard format.
        x_translation (int): The translation in the x-direction.
        y_translation (int): The translation in the y-direction.

    Raises:
        ValueError: If the JSON object type is not supported.

    Returns:
        dict: A translated JSON object.

    Example::

        json_object = {
            "type": "ellipse",
            "center": {"x": 100, "y": 200},
            "radiusX": 50,
            "radiusY": 100,
            "angleOfRotation": 0,
            "id": "12345-abcde"
        }
        x_translation = 100
        y_translation = -20

        result = {
            "type": "ellipse",
            "center": {"x": 200, "y": 180},
            "radiusX": 50,
            "radiusY": 100,
            "angleOfRotation": 0,
            "id": "12345-abcde"
        }
    """
    newjson_object = json_object.copy()
    if json_object["type"] == "ellipse":
        newjson_object["center"] = translate_coordinates(
            json_object["center"], x_translation, y_translation
        )
    elif json_object["type"] == "polygon":
        newjson_object["points"] = [
            translate_coordinates(point, x_translation, y_translation)
            for point in json_object["points"]
        ]
        if "inner_points" in json_object:
            newjson_object["inner_points"] = [
                [
                    translate_coordinates(point, x_translation, y_translation)
                    for point in inner_points
                ]
                for inner_points in json_object["inner_points"]
            ]
    else:
        raise ValueError(f"type {json_object['type']} is not supported")

    return newjson_object


def translate_coordinates(point: dict, x_translation: int, y_translation: int) -> dict:
    """Translate 2D coordinates by adding specified translations.

    Args:
        point (dict): A dictionary representing 2D coordinates with 'x' and 'y' keys.
        x_translation (int): The amount to translate in the x-axis.
        y_translation (int): The amount to translate in the y-axis.

    Returns:
        dict: A new dictionary with translated coordinates.
    """
    return {
        "x": int(round(point["x"] + x_translation)),
        "y": int(round(point["y"] + y_translation)),
    }



from __future__ import absolute_import
from unittest import TestCase
import numpy as np
from .main import fluo_tophat


class TestIlluminationCorretionAlgorithm(TestCase):
    def test_grey_image(self):
        img = np.ones((1000, 1000)) * 128
        illum_img, _, _ = fluo_tophat(img)
        self.assertEqual(np.min(illum_img), 0)

    def test_white_image(self):
        img = np.ones((1000, 1000)) * 255
        illum_img, _, _ = fluo_tophat(img)
        self.assertEqual(np.min(illum_img), 0)

    def test_black_image(self):
        img = np.ones((1000, 1000)) * 0
        illum_img, _, _ = fluo_tophat(img)
        self.assertEqual(np.min(illum_img), 0)

    def test_random_image(self):
        img = np.random.randint(0, 255, (1000, 1000))
        img[0:400, 0:400] = np.zeros((400, 400))
        illum_img, _, _ = fluo_tophat(img)
        self.assertGreaterEqual(np.min(illum_img), 0)
        self.assertLessEqual(np.max(illum_img), 255)



import cv2
import json

from tkinter import filedialog

from tomni.transformers import json2mask

file = filedialog.askopenfilename()
image_path = filedialog.askopenfilename()

with open(file, "r") as f:
    json_objects = json.load(f)

image = cv2.imread(image_path, 0)

mask = json2mask(json_objects, image.shape)

cv2.imwrite(file.replace(".json", "_mask.png"), mask * 255)



import numpy as np
import cv2
import warnings


def absolute_difference(
    img: np.ndarray, gauss_size: int = 21, new_median: int = 128
) -> np.ndarray:
    """
    Remove large blurry artifacts from an image by taking the absolute difference
    between the blurred and original image.

    Args:
        img (numpy.ndarray): The input image.
        gauss_size (int, optional): The size of the Gaussian filter kernel for blurring.
                                    Must be an odd number. Defaults to 21.
        new_median (int, optional): The target median value for the resulting image.
                                    Defaults to 128.

    Returns:
        numpy.ndarray: The processed image with blurry artifacts removed.

    Warning:
        - This function is not suitable for human vision and may produce artifacts.
        - The `gauss_size` should be an odd number; if it's even, it will be incremented by 1.

    Note:
        This algorithm is fractal, meaning that applying illumination correction
        followed by cropping will yield the same result as cropping first.
    """

    img = np.array(img, dtype=np.int16)

    if gauss_size % 2 == 0:
        gauss_size += 1  # Only works with odd gauss size
        warnings.warn(
            "the gauss_size needs to be odd. It is {}. One is added to it".format(
                gauss_size
            ),
            SyntaxWarning,
        )

    """
    The illumination is every detail bigger than the gauss_size.
    This includes the inside of square bigger than gauss_size, but not its borders.
    """
    img -= cv2.GaussianBlur(img, (gauss_size, gauss_size), 0)

    img += new_median

    # All overhead needs to be rounded down
    img = np.clip(img, 0, 255)
    return img.astype(np.uint8)



import math

import cv2
import numpy as np

from ...transformers import json2contours


def add_area(json_object: dict) -> None:
    """
    Calculate and add the area of a geometric object to a JSON annotation object.

    This function takes a JSON object representing a geometric annotation in standard AxionBio format (ellipse, polygon, or circle),
    calculates the area of the object, and adds it as the "area" key in the JSON object.

    Args:
        json_object (dict): A JSON object following the standard AxionBio format.

    Raises:
        ValueError: If the type of the annotation in the JSON object is not supported.

    Example::

        circle_json = {
            "type": "circle",
            "center": {"x": 10, "y": 10},
            "radiusX": 5
        }
        add_area(circle_json)
        print(circle_json)
        {
            "type": "circle",
            "center": {"x": 10, "y": 10},
            "radiusX": 5,
            "area": 78.53981633974483
        }

    Note:
        - Supported annotation types are "circle," "ellipse," and "polygon."
        - The area is calculated based on the properties of the geometric object:
          - For circles, it uses the formula: π * (radiusX^2)
          - For ellipses, it uses the formula: π * radiusX * radiusY
          - For polygons, it calculates the area using OpenCV's `cv2.contourArea` function.
    """
    type_anno = json_object["type"]

    if type_anno == "circle":
        area = math.pi * (json_object["radiusX"] ** 2)

    elif type_anno == "ellipse":
        area = math.pi * json_object["radiusX"] * json_object["radiusY"]

    elif type_anno == "polygon":
        outer_contour = json2contours(json_object)
        area = cv2.contourArea(outer_contour)
        if "inner_points" in json_object:
            for inner_contour in json_object["inner_points"]:
                inner_contour_result = np.array(
                    [[i["x"], i["y"]] for i in inner_contour], dtype=np.int32
                )
                inner_area = cv2.contourArea(inner_contour_result)
                area -= inner_area
    else:
        raise ValueError(f"{type_anno} is not supported")

    json_object["area"] = area



import numpy as np
import cv2


def positions2contour(
    positions: np.ndarray,
    simplify_error: float = 0,
    return_inner_contours: bool = False,
) -> np.ndarray:
    """
    Transform a list of positions into an OpenCV contour format.

    Args:
        positions (np.ndarray): A 2-dimensional array with shape [N, 2], representing the x and y coordinates
            of N pixels in the shape.
        simplify_error (float, optional): ⚠️ Deprecated parameter. No longer used.
        return_inner_contours (bool, optional): If True, return the internal contours (holes within the shape).
            If False (default), return only the external contour (the outline of the shape).

    Raises:
        DeprecationWarning: The `simplify_error` parameter is deprecated and no longer supported.

    Returns:
        Union[np.ndarray, Tuple[np.ndarray, List[np.ndarray]]]: If `return_inner_contours` is False, returns the
        external contour as an OpenCV-style np.ndarray.
        If `return_inner_contours` is True and there are internal contours, returns a tuple containing the
        external contour as the first element and a list of internal contours as the second element.
        Each contour is represented as an np.ndarray of shape [M, 2], where M is the number of points in the contour.

    Note:
        - The `simplify_error` parameter has been deprecated and is no longer used. Please update your code accordingly.
        - The input `positions` should be a 2D array with two columns, where each row represents the (x, y) coordinates
          of a pixel in the shape.
        - The `return_inner_contours` parameter controls whether internal contours (holes within the shape) are included.
          If set to True, both external and internal contours are returned as a tuple.
    """

    if simplify_error > 0:
        raise DeprecationWarning("As of Tomni 1.10 simplify is no longer supported")

    max_x = np.max(positions[:, 0])
    min_x = np.min(positions[:, 0])
    max_y = np.max(positions[:, 1])
    min_y = np.min(positions[:, 1])

    positions -= np.array([min_x, min_y])
    transposed_positions = positions.transpose()

    array_mask = np.zeros((max_y + 1 - min_y, max_x + 1 - min_x), dtype="uint8")
    array_mask[transposed_positions[1], transposed_positions[0]] = 255

    # Chose between RETR_CCOMP to retrieve all of the contours and organizes them into a two-level hierarchy
    # Or RETR_EXTERNAL retrieves only the extreme outer contours
    mode = cv2.RETR_CCOMP if return_inner_contours else cv2.RETR_EXTERNAL
    contours = cv2.findContours(array_mask, mode, cv2.CHAIN_APPROX_SIMPLE)[0]

    external_contour = contours[0] + [min_x, min_y]

    if not return_inner_contours:
        return external_contour

    if len(contours) == 1:
        internal_contours = []
    else:
        internal_contours = [i + [min_x, min_y] for i in contours[1:]]

    return external_contour, internal_contours



from __future__ import absolute_import
from unittest import TestCase
import numpy as np
from .main import relative_difference


class TestIlluminationCorretionRatioBased(TestCase):
    def test_grey_image(self):
        img = np.ones((1000, 1000)) * 128
        illum_img = relative_difference(img)
        self.assertGreaterEqual(np.min(illum_img), 0)
        self.assertLessEqual(np.max(illum_img), 255)

    def test_white_image(self):
        img = np.ones((1000, 1000)) * 255
        illum_img = relative_difference(img)
        self.assertGreaterEqual(np.min(illum_img), 0)
        self.assertLessEqual(np.max(illum_img), 255)

    def test_black_image(self):
        img = np.ones((1000, 1000)) * 0
        illum_img = relative_difference(img)
        self.assertGreaterEqual(np.min(illum_img), 0)
        self.assertLessEqual(np.max(illum_img), 255)

    def test_illumination_black_white(self):
        # BLACK GREY WHITE TEST
        xx, yy = np.mgrid[:300, :300]
        dist = xx
        img = (dist >= 100) * 128 + (dist >= 200) * 128
        illum_img = relative_difference(img, gauss_size=21)
        self.assertGreaterEqual(np.min(illum_img), 0)
        self.assertLessEqual(np.max(illum_img), 255)

    def test_illumination_3_shades_of_grey(self):
        print("\nSTART 3 SHADES OF GREY TEST")
        xx, yy = np.mgrid[:300, :300]
        dist = xx
        img = (dist >= 100) * 50 + (dist >= 200) * 50 + 50
        illum_img = relative_difference(img, gauss_size=21)
        self.assertGreaterEqual(np.min(illum_img), 0)
        self.assertLessEqual(np.max(illum_img), 255)

    def test_grey_image_adj(self):
        img = np.ones((1000, 1000)) * 128
        illum_img = relative_difference(
            img, smooth_size=3, do_normalize=True, resize_ratio=0.125
        )
        self.assertGreaterEqual(np.min(illum_img), 0)
        self.assertLessEqual(np.max(illum_img), 255)

    def test_white_image_adj(self):
        img = np.ones((1000, 1000)) * 255
        illum_img = relative_difference(
            img, smooth_size=3, do_normalize=True, resize_ratio=0.125
        )
        self.assertGreaterEqual(np.min(illum_img), 0)
        self.assertLessEqual(np.max(illum_img), 255)

    def test_black_image_adj(self):
        img = np.ones((1000, 1000)) * 0
        illum_img = relative_difference(
            img, smooth_size=3, do_normalize=True, resize_ratio=0.125
        )
        self.assertGreaterEqual(np.min(illum_img), 0)
        self.assertLessEqual(np.max(illum_img), 255)

    def test_illumination_black_white_adj(self):
        # BLACK GREY WHITE TEST
        xx, yy = np.mgrid[:300, :300]
        dist = xx
        img = (dist >= 100) * 128 + (dist >= 200) * 128
        illum_img = relative_difference(
            img, gauss_size=21, smooth_size=3, do_normalize=True, resize_ratio=0.125
        )
        self.assertGreaterEqual(np.min(illum_img), 0)
        self.assertLessEqual(np.max(illum_img), 255)

    def test_illumination_3_shades_of_grey_adj(self):
        print("\nSTART 3 SHADES OF GREY TEST")
        xx, yy = np.mgrid[:300, :300]
        dist = xx
        img = (dist >= 100) * 50 + (dist >= 200) * 50 + 50
        illum_img = relative_difference(
            img, gauss_size=21, smooth_size=3, do_normalize=True, resize_ratio=0.125
        )
        self.assertGreaterEqual(np.min(illum_img), 0)
        self.assertLessEqual(np.max(illum_img), 255)



def scale_json(json_object: dict, scaling_factor: float):
    """
    Scale a JSON object (ellipse or polygon) by a specified scaling factor.

    Args:
        json_object (dict): JSON object following the standard format (ellipse or polygon).
        scaling_factor (float): The multiplier to scale the object. A value of 1.0 keeps the object unchanged.

    Returns:
        dict: A new JSON object scaled by the provided factor.

    Raises:
        ValueError: If the JSON object type is not supported.

    Example::

        json_object = {
            "type": "ellipse",
            "center": {"x": 100, "y": 200},
            "radiusX": 50,
            "radiusY": 100,
            "angleOfRotation": 0,
            "id": "12345-abcde"}
        scaling_factor = 2

        result = {
            "type": "ellipse",
            "center": {"x": 200, "y": 400},
            "radiusX": 100,
            "radiusY": 200,
            "angleOfRotation": 0,
            "id": "12345-abcde"}
    """
    newjson_object = json_object.copy()
    if json_object["type"] == "ellipse":
        newjson_object["radiusX"] = int(round(json_object["radiusX"] * scaling_factor))
        newjson_object["radiusY"] = int(round(json_object["radiusY"] * scaling_factor))
        newjson_object["center"] = scale_coordinates(
            json_object["center"], scaling_factor
        )

    elif json_object["type"] == "polygon":
        newjson_object["points"] = [
            scale_coordinates(point, scaling_factor) for point in json_object["points"]
        ]
        if "inner_points" in json_object:
            newjson_object["inner_points"] = [
                [
                    scale_coordinates(inner_point, scaling_factor)
                    for inner_point in inner_points
                ]
                for inner_points in json_object["inner_points"]
            ]
    else:
        raise ValueError(f"type {json_object['type']} is not supported")

    return newjson_object


def scale_coordinates(point: dict, scaling_factor: float) -> dict:
    """Scale a 2D point by a specified scaling factor.

    Args:
        point (dict): A dictionary representing the 2D point with 'x' and 'y' coordinates.
        scaling_factor (float): The multiplier to scale the point.

    Returns:
        dict: A new dictionary representing the scaled 2D point.
    """
    return {
        "x": int(round(point["x"] * scaling_factor)),
        "y": int(round(point["y"] * scaling_factor)),
    }



from typing import List
from copy import deepcopy
import numpy as np

from ...transformers import json2dict
from .. import add_area, add_circularity


def summary_json(
    json_list: List[dict], keyword: str, do_copy: bool = True, rounding: int = 2
) -> tuple:
    """
    Create a summary of a specified keyword variable inside JSON objects.

    Args:
        json_list (List[dict]): A list of JSON objects in standard AxionBio format.
        keyword (str): The keyword for which to create the summary (e.g., 'area').
        do_copy (bool, optional): When True, copies the JSON list and does not modify it. Defaults to True.
        rounding (int, optional): The number of decimals used in the summary. Defaults to 2.

    Raises:
        ValueError: If an unsupported keyword is provided.

    Returns:
        tuple: A tuple containing:
            - Number of objects
            - Total sum of the keyword in objects
            - Mean value of the keyword in objects
            - Maximum value of the keyword in objects
            - Minimum value of the keyword in objects

    Example::

        json_list = [
            {"area": 10},
            {"area": 15},
            {"area": 20},
        ]
        keyword = "area"

        output = summary_json(json_list, keyword)

        result = (3, 45, 15.0, 20, 10)
    """

    if do_copy:
        copy_json_list = deepcopy(json_list)
    else:
        copy_json_list = json_list

    if keyword == "area":
        if not keyword in json_list[0]:
            for json_object in copy_json_list:
                add_area(json_object)
    else:
        raise ValueError(f"summary for {keyword} is not supported")

    keyword_list = json2dict(copy_json_list, [keyword])[keyword]

    return (
        len(keyword_list),  # Number of objects
        round(np.sum(keyword_list), rounding),  # Total sum of objects
        round(np.mean(keyword_list), rounding),  # Mean of objects
        round(np.max(keyword_list), rounding),  # Max of objects
        round(np.min(keyword_list), rounding),  # Min of objects
    )



from typing import Union
import numpy as np
import cv2


def convert_color(img: np.ndarray, new_color: Union[str, int]) -> np.ndarray:
    """
    Convert the color format of an image to the desired color space (GRAY, BGR, or BGRA).

    Args:
        img (np.ndarray): The image to be converted. Supported color spaces: GRAY (1), BGR (3), and BGRA (4).
        new_color (Union[str, int]): The desired color space for the converted image.

    Raises:
        ValueError: If new_color is not supported.
        ValueError: If the input image has an unsupported number of dimensions.
        ValueError: If the input image has an unsupported number of color channels.

    Returns:
        np.ndarray: The converted image.

    Note:
        Supported values:

        - (1) 'GRAY'/'GREY'
        - (3) 'BGR'/'COLOR'/'COLOUR'
        - (4) 'BGRA'/'TRANSPARENT'

    """

    img = img.astype(np.uint8)
    # Not getting into that debate
    if type(new_color) is str:
        if new_color.upper() in ["GRAY", "GREY"]:
            new_color = 1
        # Also not getting into that debate
        elif new_color.upper() in ["BGR", "COLOR", "COLOUR"]:
            new_color = 3
        elif new_color.upper() in ["BGRA", "TRANSPARENT"]:
            new_color = 4
        else:
            new_color = -1

    if new_color not in [1, 3, 4]:
        raise ValueError(
            "The image type {} is not supported. Currently supported: GRAY (1), BGR (3) and BGRA (4)".format(
                new_color
            )
        )

    # Determine the color type of current image
    if len(img.shape) == 2:
        currentType = 1
    elif len(img.shape) != 3:
        raise ValueError(
            "The image does not have known dimensions ({}). It should have 2 or 3 dimensions.".format(
                img.shape
            )
        )
    elif img.shape[2] == 1:
        img = img[:, :, 0]
        currentType = 1
    elif img.shape[2] == 3:
        currentType = 3
    elif img.shape[2] == 4:
        currentType = 4
    else:
        raise ValueError(
            "The image type {} is not supported. Currently supported: GRAY (1), BGR (3) and BGRA (4)".format(
                img.shape[2]
            )
        )

    # That was easy
    if currentType == new_color:
        return img

    if currentType == 1 and new_color == 3:
        return cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    if currentType == 1 and new_color == 4:
        return cv2.cvtColor(img, cv2.COLOR_GRAY2BGRA)

    if currentType == 3 and new_color == 1:
        return cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    if currentType == 3 and new_color == 4:
        return cv2.cvtColor(img, cv2.COLOR_BGR2BGRA)

    if currentType == 4 and new_color == 1:
        return cv2.cvtColor(img, cv2.COLOR_BGRA2GRAY)
    if currentType == 4 and new_color == 3:
        return cv2.cvtColor(img, cv2.COLOR_BGRA2BGR)



import unittest
from .main import check_overlap_bbox


class TestOverlap(unittest.TestCase):
    # (xmin, ymin, xmax, ymax)
    def test_overlap_x(self):
        bb1 = [0, 0, 10, 10]
        bb2 = [5, 0, 15, 10]
        output = check_overlap_bbox(bb1, bb2)
        self.assertEqual(output, True)

    def test_overlap_y(self):
        bb1 = [0, 0, 10, 10]
        bb2 = [0, 5, 10, 15]
        output = check_overlap_bbox(bb1, bb2)
        self.assertEqual(output, True)

    def test_overlap_xy(self):
        bb1 = [0, 0, 10, 10]
        bb2 = [5, 5, 15, 15]
        output = check_overlap_bbox(bb1, bb2)
        self.assertEqual(output, True)

    def test_touch_x(self):
        bb1 = [0, 0, 10, 10]
        bb2 = [10, 0, 15, 10]
        output = check_overlap_bbox(bb1, bb2)
        self.assertEqual(output, False)

    def test_touch_y(self):
        bb1 = [0, 0, 10, 10]
        bb2 = [0, 10, 10, 15]
        output = check_overlap_bbox(bb1, bb2)
        self.assertEqual(output, False)

    def test_touch_xy(self):
        bb1 = [0, 0, 10, 10]
        bb2 = [10, 10, 15, 15]
        output = check_overlap_bbox(bb1, bb2)
        self.assertEqual(output, False)

    def test_complete_miss(self):
        bb1 = [0, 0, 10, 10]
        bb2 = [20, 20, 30, 30]
        output = check_overlap_bbox(bb1, bb2)
        self.assertEqual(output, False)

    def test_wrong_length(self):
        bb1 = [0, 10, 0, 10, 2]
        bb2 = [10, 15, 10, 15, 2]
        with self.assertRaises(ValueError):
            check_overlap_bbox(bb1, bb2)

    def test_wrong_length_one(self):
        bb1 = [0, 10, 0, 10, 2]
        bb2 = [10, 15, 10, 15]
        with self.assertRaises(ValueError):
            check_overlap_bbox(bb1, bb2)



import uuid
from typing import Union


def ellipse2json(
    x: int, y: int, radius: int, secondRadius: Union[None, int] = None, alpha: float = 0
) -> dict:
    """
    Convert ellipse parameters into a JSON object in standard AxionBio format.

    Args:
        x (int): The center x-position of the ellipse.
        y (int): The center y-position of the ellipse.
        radius (int): The radius of the ellipse.
        secondRadius (int, None, optional): The smallest radius of the ellipse. Defaults to None.
            In case of a circle or when not provided, it is assumed to be equal to 'radius'.
        alpha (float, optional): This is the angle between the biggest radius and the x-axis.
            Defaults to 0.

    Returns:
        dict: A JSON object representing the ellipse in standard AxionBio format.
    """
    result = []
    result = {
        "type": "ellipse",
        "center": {"x": int(x), "y": int(y)},
        "radiusX": int(radius),
        "radiusY": int(secondRadius) if secondRadius else int(radius),
        "angleOfRotation": alpha,
        "id": str(uuid.uuid4()),
    }
    return result



import numpy as np


def img_dim(arr, showChannels=False):
    """
    Get the dimensions and optionally the number of color channels of a numpy array representing an image.

    Args:
        arr (numpy.ndarray): The image represented as a numpy array.
        showChannels (bool, optional): Whether to include the number of color channels in the output.
                                        True to show channels, False to hide them. Defaults to False.

    Returns:
        tuple: A tuple containing the dimensions of the image (width, height) or (width, height, channels)
                depending on the value of showChannels.

    Example::

        import numpy as np
        image = np.zeros((100, 200, 3), dtype=np.uint8)
        dimensions = img_dim(image)
        print(dimensions)  # Output: (200, 100)

        dimensions_with_channels = img_dim(image, showChannels=True)
        print(dimensions_with_channels)  # Output: (200, 100, 3)
    """
    s = np.shape(arr)

    if showChannels:
        if len(s) > 2:
            return s[1], s[0], s[2]
        else:
            return s[1], s[0], 1
    else:
        return s[1], s[0]



from ...transformers import json2contours
from ...contour_operations import get_center


def add_center(json_object: dict):
    """
    Calculate and add the center coordinates to a JSON annotation object.

    This function takes a JSON object representing an annotation in standard AxionBio format (ellipse or polygon),
    calculates the center coordinates of the object, and adds them as the "center" key in the JSON object.

    Args:
        json_object (dict): A JSON object following the standard AxionBio format.

    Raises:
        ValueError: If the type of the annotation in the JSON object is not supported.

    Example::

        # Define a sample JSON annotation object in AxionBio format (polygon)
        sample_json_object = {
            "type": "polygon",
            "points": [[100, 100], [200, 100], [200, 200], [100, 200]]
        }

        # Call the add_center function to calculate and add the center coordinates
        add_center(sample_json_object)

        # The updated JSON object will contain the added "center" key
        # Updated JSON Object:
        updated_json_object = {
            'type': 'polygon',
            'points': [[100, 100], [200, 100], [200, 200], [100, 200]],
            'center': {'x': 150.0, 'y': 150.0}
        }

    Note:
        - Supported annotation types are "ellipse" and "polygon."
        - For ellipses, no action is taken as the center is already included in the JSON object.
        - For polygons, the center coordinates are calculated using the `get_center` function from the `contour_operations` module.
    """
    if json_object["type"] == "ellipse":
        pass
    elif json_object["type"] == "polygon":
        contour = json2contours(json_object)
        x, y = get_center(contour)
        json_object["center"] = {"x": x, "y": y}
    else:
        raise ValueError(f"No json type found of {json_object['type']}")



from unittest import TestCase

import numpy as np

from tomni.annotation_manager.annotations.point.main import Point

from .main import compress_polygon_points


class TestCompressPolygonPoints(TestCase):
    def setUp(self) -> None:
        self.points = [
            Point(1, 2),
            Point(1, 4),
            Point(2, 5),
            Point(4, 5),
            Point(5, 4),
            Point(5, 2),
            Point(4, 1),
            Point(3, 1),
            Point(2, 1),
        ]

    def test_compress_happy_flow(self):
        expected = [
            Point(x=1, y=2),
            Point(x=2, y=5),
            Point(x=5, y=4),
            Point(x=4, y=1),
            Point(x=2, y=1),
        ]
        actual = compress_polygon_points(self.points, 1)

        self.assertEqual(expected, actual)

    def test_compress_to_much_so_return_same(self):
        expected = [
            Point(1, 2),
            Point(1, 4),
            Point(2, 5),
            Point(4, 5),
            Point(5, 4),
            Point(5, 2),
            Point(4, 1),
            Point(3, 1),
            Point(2, 1),
        ]
        actual = compress_polygon_points(self.points, 3)

        self.assertEqual(expected, actual)



import cv2
import numpy as np
from scipy.stats import kurtosis, skew


def rescale_intensity(img, min_val, max_val):
    """
    Rescale pixel intensities of an image to the specified minimum and maximum values.

    Args:
        img (numpy.ndarray): The input image.
        min_val (int): The minimum pixel intensity value.
        max_val (int): The maximum pixel intensity value.

    Returns:
        numpy.ndarray: The rescaled image.
    """
    output_img = np.clip(img, min_val, max_val)
    output_img = (output_img - min_val) / (max_val - min_val) * 255
    return output_img.astype(np.uint8)


def fluo_tophat(img, normalize=0):
    """
    Fluorescent Top-Hat Background Subtraction

    This function performs background subtraction on a fluorescent channel image using a morphological white top-hat filter.
    The white top-hat filter is effective at highlighting small bright structures against a darker background.

    Args:
        img (numpy.ndarray): A grayscale image where background subtraction will be applied.
        normalize (int, optional): If set to 1, the resulting image will be normalized to the 0-255 range. Defaults to 0.

    Returns:
        img (numpy.ndarray): The image after background subtraction.
        p2 (float): The lower bound value used for potential later normalization.
        p98 (float): The upper bound value used for potential later normalization.

    Note:
        The `p2` and `p98` values represent the lower and upper bounds for potential image normalization. They are not automatically applied to the image to avoid data loss. These values can be used during integration and passed to the front-end for further processing.

    Example:
        To perform background subtraction on an image and normalize the result, you can use this function as follows:

        >>> result_image, lower_bound, upper_bound = fluo_tophat(image, normalize=1)
    """

    img = img.astype(np.uint8)
    img = cv2.GaussianBlur(img, (5, 5), 0)
    filterSize = (25, 25)
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, filterSize)
    img = cv2.morphologyEx(img, cv2.MORPH_TOPHAT, kernel, iterations=10)

    kur = np.mean(kurtosis(img, fisher=True))
    skew1 = np.mean(skew(img))
    if kur > 1 and skew1 > 1:
        p2, p98 = np.percentile(img, (15, 99.5), interpolation="linear")
        if normalize == 1:
            img = rescale_intensity(img, p2, p98)
    else:
        p2, p98 = np.percentile(img, (15, 100), interpolation="linear")
        if normalize == 1:
            img = rescale_intensity(img, p2, p98)

    return img, p2, p98



import numpy as np

from ...img_dim import img_dim
from ...make_mask import make_mask_contour, make_mask_ellipse
from ..json2contours import json2contours


def contour2mask(mask: np.ndarray, contour: list) -> np.ndarray:
    """Converts contour to mask"""
    add_mask = make_mask_contour(img_dim(mask), contour)
    mask = mask + add_mask
    mask[mask > 1] = 1
    return mask


def ellipse2mask(mask: np.ndarray, x: int, y: int, r1: float, r2: float) -> np.ndarray:
    """Converts ellipse to mask"""
    add_mask = make_mask_ellipse(img_dim(mask), x, y, r1, r2)
    mask = mask + add_mask
    mask[mask > 1] = 1
    return mask


def json2mask(
    json_objects: list, img_shape: tuple, minimum_size_contours: int = 3
) -> np.ndarray:
    """
    Convert a list of JSON objects in standard AxionBio format to a binary mask.

    Args:
        json_objects (list): A list of JSON objects with annotations in the standard AxionBio format.
        img_shape (tuple): The dimensions (height, width) of the mask.
        minimum_size_contours (int, optional): The minimum number of points a contour should have to be included.
            Defaults to 3. Set to 0 to include all contours regardless of size.

    Returns:
        np.ndarray: A binary mask as a NumPy array.

    Raises:
        TypeError: If a JSON object of an unsupported type is encountered.

    Example::

        json_objects = [
            {
                "type": "polygon",
                "points": [
                    {"x": 0, "y": 0},
                    {"x": 0, "y": 5},
                    {"x": 5, "y": 5},
                    {"x": 5, "y": 0},
                ],
            }
        ]
        img_shape = (10, 10)
        minimum_size_contours = 3
        result = json2mask(json_objects, img_shape, minimum_size_contours)
        print(result)
        [[1 1 1 1 1 0 0 0 0 0]
         [1 1 1 1 1 0 0 0 0 0]
         [1 1 1 1 1 0 0 0 0 0]
         [1 1 1 1 1 0 0 0 0 0]
         [1 1 1 1 1 0 0 0 0 0]
         [0 0 0 0 0 0 0 0 0 0]
         [0 0 0 0 0 0 0 0 0 0]
         [0 0 0 0 0 0 0 0 0 0]
         [0 0 0 0 0 0 0 0 0 0]
         [0 0 0 0 0 0 0 0 0 0]]
    """
    mask = np.zeros(img_shape, dtype=np.uint8)
    for json_object in json_objects:
        if json_object["type"] == "polygon":
            if len(json_object["points"]) >= minimum_size_contours:
                contour = json2contours(json_object)
                contour = [x[0] for x in contour]
                mask = contour2mask(mask, contour)
        elif json_object["type"] == "ellipse":
            x = json_object["center"]["x"]
            y = json_object["center"]["y"]
            r1 = json_object["radiusX"]
            r2 = json_object["radiusY"]
            mask = ellipse2mask(mask, x, y, r1, r2)
        else:
            raise TypeError(
                f"json object {json_object['type']} could not be converted to a mask"
            )

    return mask



from unittest import TestCase

from .main import flip_json


class TestFlipPatch(TestCase):
    def test_flips(self):
        input_json = {
            "type": "ellipse",
            "center": {"x": 9, "y": 3},
            "radiusX": 5,
            "radiusY": 5,
            "angleOfRotation": 0,
        }

        result_json = flip_json(input_json, 10)
        expected_json = {
            "type": "ellipse",
            "center": {"x": 9, "y": 6},
            "radiusX": 5,
            "radiusY": 5,
            "angleOfRotation": 0,
        }

        self.assertDictEqual(result_json, expected_json)

    def test_flips_polygon(self):
        input_json = {
            "type": "polygon",
            "points": [{"x": 7, "y": 2}, {"x": 7, "y": 6}, {"x": 12, "y": 5}],
        }
        result_json = flip_json(input_json, 10)
        expected_json = {
            "type": "polygon",
            "points": [{"x": 7, "y": 7}, {"x": 7, "y": 3}, {"x": 12, "y": 4}],
        }
        self.assertDictEqual(result_json, expected_json)

    def test_flips_polygon_inner_points(self):
        input_json = {
            "type": "polygon",
            "points": [{"x": 7, "y": 2}, {"x": 7, "y": 6}, {"x": 12, "y": 5}],
            "inner_points": [
                [{"x": 1, "y": 1}, {"x": 1, "y": 5}],
                [{"x": 2, "y": 1}, {"x": 1, "y": 2}],
            ],
        }
        result_json = flip_json(input_json, 10)
        expected_json = {
            "type": "polygon",
            "points": [{"x": 7, "y": 7}, {"x": 7, "y": 3}, {"x": 12, "y": 4}],
            "inner_points": [
                [{"x": 1, "y": 8}, {"x": 1, "y": 4}],
                [{"x": 2, "y": 8}, {"x": 1, "y": 7}],
            ],
        }
        self.assertDictEqual(result_json, expected_json)

    def test_flips_polygon_inner_points_empty(self):
        input_json = {
            "type": "polygon",
            "points": [{"x": 7, "y": 2}, {"x": 7, "y": 6}, {"x": 12, "y": 5}],
            "inner_points": [],
        }
        result_json = flip_json(input_json, 10)
        expected_json = {
            "type": "polygon",
            "points": [{"x": 7, "y": 7}, {"x": 7, "y": 3}, {"x": 12, "y": 4}],
            "inner_points": [],
        }
        self.assertDictEqual(result_json, expected_json)



from re import S
from typing import Tuple, Union

import numpy as np

from tomni.annotation_manager.annotations.point import Point
from tomni.annotation_manager.annotations.annotation import Annotation


class BinaryMask(Annotation):
    def __init__(self, binary_mask: np.ndarray):
        # always on [0,1]? if not parse. or allow multiple annotations in mask?
        self._binary_mask = binary_mask
        pass

    @property
    def get_mask(self) -> np.ndarray:
        return self._binary_mask

    def get_shape(self) -> Union[Tuple[int, int], Tuple[int, int, int]]:
        """_summary_

        Returns:
            Tuple[int, int]: _description_
        """
        pass

    def get_feature_x(self) -> None:
        """
        not sure what mask features we commonly use but they could be here.
        """
        pass



from typing import Tuple

import cv2
import numpy as np


def get_center(contour: np.ndarray) -> Tuple[int, int]:
    """
    Calculate the center coordinates of a contour.

    Args:
        contour (np.ndarray): An OpenCV contour of a single object.

    Returns:
        Tuple[int, int]: A tuple containing the x and y coordinates of the center.

    Example::

        contour = np.array([[[1, 2]], [[2, 3]], [[3, 2]], [[2, 1]]])
        get_center(contour)
        (2, 2)
    """
    M = cv2.moments(contour)
    if M["m00"] == 0:
        return contour[0][0][0], contour[0][0][1]
    cX = int(M["m10"] / M["m00"])
    cY = int(M["m01"] / M["m00"])

    return cX, cY



import cv2
import numpy as np


def bbox_fitting(
    img: np.ndarray, x1: int, y1: int, x2: int, y2: int, padding_value: int = 0
) -> np.ndarray:
    """
    This function extracts a region of interest (ROI) from an input image based on a specified bounding box.
    It is particularly useful for isolating specific areas of interest within an image.

    Parameters:
        - img (np.ndarray): The input image from which the bounding box will be extracted.
        - x1 (int): The lowest value of X-coordinate for the bounding box.
        - y1 (int): The lowest value of Y-coordinate for the bounding box.
        - x2 (int): The highest value of X-coordinate for the bounding box.
        - y2 (int): The highest value of Y-coordinate for the bounding box.
        - padding_value (int, optional): The value assigned to pixels outside the image but within the bounding box. Defaults to 0.

    Returns:
        - np.ndarray: The extracted image patch within the specified bounding box.

    Raises:
        - TypeError: If img is not a numpy.ndarray.
        - ValueError: If x1, y1, x2, or y2 are not positive integers.

    Note:
        The coordinates used in this function are in IMAGE coordinates, which follow the convention used by Pillow and OpenCV but are swapped compared to the Numpy and Scipy convention (ARRAY coordinates). Please be mindful of this difference.

    """

    if not isinstance(img, np.ndarray):
        raise TypeError("Img needs to be a numpy.ndarry not {}".format(type(img)))

    if (
        not isinstance(x1, int)
        or not isinstance(x2, int)
        or not isinstance(y1, int)
        or not isinstance(y2, int)
    ):
        raise ValueError(
            "x1, y1, x2 and y2 need to be a positive int not {}, {}, {}, {}".format(
                type(x1), type(y1), type(x2), type(y2)
            )
        )

    if x2 < 0:
        return np.zeros((y2 - y1, x2 - x1))
    if y2 < 0:
        return np.zeros((y2 - y1, x2 - x1))
    if x1 >= img.shape[1]:
        return np.zeros((y2 - y1, x2 - x1))
    if y1 >= img.shape[0]:
        return np.zeros((y2 - y1, x2 - x1))

    cy1 = max(0, y1)
    cy2 = min(img.shape[0], y2)
    cx1 = max(0, x1)
    cx2 = min(img.shape[1], x2)

    pad_left = -min(0, x1)
    pad_right = max(x2 - img.shape[1], 0)
    pad_top = -min(0, y1)
    pad_bottom = max(y2 - img.shape[0], 0)

    img = img[cy1:cy2, cx1:cx2]

    img = cv2.copyMakeBorder(
        img,
        top=pad_top,
        bottom=pad_bottom,
        left=pad_left,
        right=pad_right,
        borderType=cv2.BORDER_CONSTANT,
        value=padding_value,
    )

    return img
